@@ -14,7 +14,7 @@
  */
 /* 
  * liblognorm - a fast samples-based log normalization library
- * Copyright 2010 by Rainer Gerhards and Adiscon GmbH.
+ * Copyright 2010-2016 by Rainer Gerhards and Adiscon GmbH.
  *
  * Modified by Pavel Levshin (pavel@levshin.spb.ru) in 2013
  *
@@ -45,9 +45,9 @@
 
 #include <libestr.h>
 
-#include "json_compatibility.h"
 #include "lognorm.h"
 #include "internal.h"
+#include "enc.h"
 
 static char hexdigit[16] =
 	{'0', '1', '2', '3', '4', '5', '6', '7', '8',
@@ -60,7 +60,7 @@ static char hexdigit[16] =
  * byte-by-byte basis, which simply is incorrect.
  * rgerhards, 2010-11-09
  */
-int
+static int
 ln_addValue_CSV(const char *buf, es_str_t **str)
 {
 	int r;
@@ -128,7 +128,7 @@ ln_addValue_CSV(const char *buf, es_str_t **str)
 }
 
 
-int
+static int
 ln_addField_CSV(struct json_object *field, es_str_t **str)
 {
 	int r, i;
@@ -159,6 +159,12 @@ ln_addField_CSV(struct json_object *field, es_str_t **str)
 		CHKN(value = json_object_get_string(field));
 		CHKR(ln_addValue_CSV(value, str));
 		break;
+	case json_type_null:
+	case json_type_boolean:
+	case json_type_double:
+	case json_type_object:
+		CHKR(es_addBuf(str, "***unsupported type***", sizeof("***unsupported type***")-1));
+		break;
 	default:
 		CHKR(es_addBuf(str, "***OBJECT***", sizeof("***OBJECT***")-1));
 	}
@@ -197,7 +203,7 @@ ln_fmtEventToCSV(struct json_object *json, es_str_t **str, es_str_t *extraData)
 			*nn = '\0';
 			nn++;
 		}
-		field = json_object_object_get(json, name);
+		json_object_object_get_ex(json, name, &field);
 		if (needComma) {
 			CHKR(es_addChar(str, ','));
 		} else {
