@@ -8,7 +8,7 @@
  */
 /* 
  * liblognorm - a fast samples-based log normalization library
- * Copyright 2010 by Rainer Gerhards and Adiscon GmbH.
+ * Copyright 2010-2016 by Rainer Gerhards and Adiscon GmbH.
  *
  * Modified by Pavel Levshin (pavel@levshin.spb.ru) in 2013
  *
@@ -39,11 +39,11 @@
 
 #include <libestr.h>
 
-#include "json_compatibility.h"
 #include "internal.h"
 #include "liblognorm.h"
+#include "enc.h"
 
-int
+static int
 ln_addValue_Syslog(const char *value, es_str_t **str)
 {
 	int r;
@@ -100,7 +100,7 @@ ln_addValue_Syslog(const char *value, es_str_t **str)
 }
 
 
-int
+static int
 ln_addField_Syslog(char *name, struct json_object *field, es_str_t **str)
 {
 	int r;
@@ -132,6 +132,12 @@ ln_addField_Syslog(char *name, struct json_object *field, es_str_t **str)
 		CHKN(value = json_object_get_string(field));
 		CHKR(ln_addValue_Syslog(value, str));
 		break;
+	case json_type_null:
+	case json_type_boolean:
+	case json_type_double:
+	case json_type_object:
+		CHKR(es_addBuf(str, "***unsupported type***", sizeof("***unsupported type***")-1));
+		break;
 	default:
 		CHKR(es_addBuf(str, "***OBJECT***", sizeof("***OBJECT***")-1));
 	}
@@ -183,14 +189,18 @@ ln_fmtEventToRFC5424(struct json_object *json, es_str_t **str)
 
 	es_addBuf(str, "[cee@115", 8);
 	
-	if((tags = json_object_object_get(json, "event.tags")) != NULL) {
+	if(json_object_object_get_ex(json, "event.tags", &tags)) {
 		CHKR(ln_addTags_Syslog(tags, str));
 	}
-	json_object_object_foreach(json, name, field) {
+	struct json_object_iterator it = json_object_iter_begin(json);
+	struct json_object_iterator itEnd = json_object_iter_end(json);
+	while (!json_object_iter_equal(&it, &itEnd)) {
+		char *const name = (char*)json_object_iter_peek_name(&it);
 		if (strcmp(name, "event.tags")) {
 			es_addChar(str, ' ');
-			ln_addField_Syslog(name, field, str);
+			ln_addField_Syslog(name, json_object_iter_peek_value(&it), str);
 		}
+		json_object_iter_next(&it);
 	}
 	es_addChar(str, ']');
 
