@@ -31,7 +31,7 @@
  *//*
  *
  * liblognorm - a fast samples-based log normalization library
- * Copyright 2010 by Rainer Gerhards and Adiscon GmbH.
+ * Copyright 2010-2016 by Rainer Gerhards and Adiscon GmbH.
  *
  * Modified by Pavel Levshin (pavel@levshin.spb.ru) in 2013
  *
@@ -58,6 +58,13 @@
 
 #include "liblognorm.h"
 
+#include <libestr.h>
+
+/* we need to turn off this warning, as it also comes up in C99 mode, which
+ * we use.
+ */
+#pragma GCC diagnostic ignored "-Wdeclaration-after-statement"
+
 /* support for simple error checking */
 
 #define CHKR(x) \
@@ -86,4 +93,13 @@ done:
 	return NULL;
 }
 
+const char * ln_DataForDisplayCharTo(__attribute__((unused)) ln_ctx ctx, void *const pdata);
+const char * ln_DataForDisplayLiteral(__attribute__((unused)) ln_ctx ctx, void *const pdata);
+const char * ln_JsonConfLiteral(__attribute__((unused)) ln_ctx ctx, void *const pdata);
+
+/* here we add some stuff from the compatibility layer */
+#ifndef HAVE_STRNDUP
+char * strndup(const char *s, size_t n);
+#endif
+
 #endif /* #ifndef INTERNAL_H_INCLUDED */
