@@ -28,13 +28,12 @@
 #include <string.h>
 #include <errno.h>
 
-#include "json_compatibility.h"
 #include "liblognorm.h"
 #include "lognorm.h"
 #include "annot.h"
 #include "samp.h"
-
-#define ERR_ABORT {r = 1; goto done; }
+#include "v1_liblognorm.h"
+#include "v1_ptree.h"
 
 #define CHECK_CTX \
 	if(ctx->objID != LN_ObjID_CTX) { \
@@ -42,22 +41,20 @@
 		goto done; \
 	}
 
-char *
+const char *
 ln_version(void)
 {
 	return VERSION;
 }
 
-ln_ctx
-ln_inherittedCtx(ln_ctx parent)
+int
+ln_hasAdvancedStats(void)
 {
-	ln_ctx child = ln_initCtx();
-	if (child != NULL) {
-		child->allowRegex = parent->allowRegex;
-		child->dbgCB = parent->dbgCB;
-		child->dbgCookie = parent->dbgCookie;
-	}
-	return child;
+#ifdef	ADVANCED_STATS
+	return 1;
+#else
+	return 0;
+#endif
 }
 
 ln_ctx
@@ -67,21 +64,27 @@ ln_initCtx(void)
 	if((ctx = calloc(1, sizeof(struct ln_ctx_s))) == NULL)
 		goto done;
 
+#ifdef HAVE_JSON_GLOBAL_SET_STRING_HASH
+	json_global_set_string_hash(JSON_C_STR_HASH_PERLLIKE);
+#endif
+#ifdef HAVE_JSON_GLOBAL_SET_PRINTBUF_INITIAL_SIZE
+	json_global_set_printbuf_initial_size(2048);
+#endif
 	ctx->objID = LN_ObjID_CTX;
 	ctx->dbgCB = NULL;
-	ctx->allowRegex = 0;
+	ctx->opts = 0;
 
 	/* we add an root for the empty word, this simplifies parse
-	 * tree handling.
+	 * dag handling.
 	 */
-	if((ctx->ptree = ln_newPTree(ctx, NULL)) == NULL) {
+	if((ctx->pdag = ln_newPDAG(ctx)) == NULL) {
 		free(ctx);
 		ctx = NULL;
 		goto done;
 	}
 	/* same for annotation set */
 	if((ctx->pas = ln_newAnnotSet(ctx)) == NULL) {
-		ln_deletePTree(ctx->ptree);
+		ln_pdagDelete(ctx->pdag);
 		free(ctx);
 		ctx = NULL;
 		goto done;
@@ -92,8 +95,8 @@ done:
 }
 
 void
-ln_setCtxOpts(ln_ctx ctx, int allow_regex) {
-	ctx->allowRegex = allow_regex;
+ln_setCtxOpts(ln_ctx ctx, const unsigned opts) {
+	ctx->opts |= opts;
 }
 
 
@@ -104,9 +107,19 @@ ln_exitCtx(ln_ctx ctx)
 
 	CHECK_CTX;
 
+	ln_dbgprintf(ctx, "exitCtx %p", ctx);
 	ctx->objID = LN_ObjID_None; /* prevent double free */
+	/* support for old cruft */
 	if(ctx->ptree != NULL)
 		ln_deletePTree(ctx->ptree);
+	/* end support for old cruft */
+	if(ctx->pdag != NULL)
+		ln_pdagDelete(ctx->pdag);
+	for(int i = 0 ; i < ctx->nTypes ; ++i) {
+		free((void*)ctx->type_pdags[i].name);
+		ln_pdagDelete(ctx->type_pdags[i].pdag);
+	}
+	free(ctx->type_pdags);
 	if(ctx->rulePrefix != NULL)
 		es_deleteStr(ctx->rulePrefix);
 	if(ctx->pas != NULL)
@@ -142,40 +155,19 @@ done:
 	return r;
 }
 
-
-int
-ln_loadSample(ln_ctx ctx, const char *buf)
-{
-    // Something bad happened - no new sample
-    if (ln_processSamp(ctx, buf, strlen(buf)) == NULL) {
-        return 1;
-    }
-    return 0;
-}
-
-
 int
 ln_loadSamples(ln_ctx ctx, const char *file)
 {
 	int r = 0;
-	FILE *repo;
-	struct ln_samp *samp;
-	int isEof = 0;
-
+	const char *tofree;
 	CHECK_CTX;
-	if(file == NULL) ERR_ABORT;
-	if((repo = fopen(file, "r")) == NULL) {
-		ln_errprintf(ctx, errno, "cannot open file %s", file);
-		ERR_ABORT;
-	}
-	while(!isEof) {
-		if((samp = ln_sampRead(ctx, repo, &isEof)) == NULL) {
-			/* TODO: what exactly to do? */
-		}
-	}
-	fclose(repo);
-
+	ctx->conf_file = tofree = strdup(file);
+	ctx->conf_ln_nbr = 0;
+	++ctx->include_level;
+	r = ln_sampLoad(ctx, file);
+	--ctx->include_level;
+	free((void*)tofree);
+	ctx->conf_file = NULL;
 done:
 	return r;
 }
-
