@@ -12,7 +12,7 @@
  *
  *//*
  * liblognorm - a fast samples-based log normalization library
- * Copyright 2010-2013 by Rainer Gerhards and Adiscon GmbH.
+ * Copyright 2010-2016 by Rainer Gerhards and Adiscon GmbH.
  *
  * This file is part of liblognorm.
  *
@@ -38,52 +38,67 @@
 #include <getopt.h>
 #include <libestr.h>
 
-#include "json_compatibility.h"
 #include "liblognorm.h"
-#include "ptree.h"
 #include "lognorm.h"
 #include "enc.h"
 
+/* we need to turn off this warning, as it also comes up in C99 mode, which
+ * we use.
+ */
+#pragma GCC diagnostic ignored "-Wdeclaration-after-statement"
+
 static ln_ctx ctx;
 
 static int verbose = 0;
-static int parsedOnly = 0;	/**< output unparsed messages? */
+#define OUTPUT_PARSED_RECS 0x01
+#define OUTPUT_UNPARSED_RECS 0x02
+static int recOutput = OUTPUT_PARSED_RECS | OUTPUT_UNPARSED_RECS; 
+				/**< controls which records to output */
+static int outputSummaryLine = 0;
+static int outputNbrUnparsed = 0;
+static int addErrLineNbr = 0;	/**< add line number info to unparsed events */
 static int flatTags = 0;	/**< print event.tags in JSON? */
 static FILE *fpDOT;
 static es_str_t *encFmt = NULL; /**< a format string for encoder use */
 static es_str_t *mandatoryTag = NULL; /**< tag which must be given so that mesg will
 					   be output. NULL=all */
-static enum { f_syslog, f_json, f_xml, f_csv } outfmt = f_syslog;
+static enum { f_syslog, f_json, f_xml, f_csv, f_raw } outfmt = f_json;
 
-void
+static void
 errCallBack(void __attribute__((unused)) *cookie, const char *msg,
 	    size_t __attribute__((unused)) lenMsg)
 {
 	fprintf(stderr, "liblognorm error: %s\n", msg);
 }
 
-void
+static void
 dbgCallBack(void __attribute__((unused)) *cookie, const char *msg,
 	    size_t __attribute__((unused)) lenMsg)
 {
 	fprintf(stderr, "liblognorm: %s\n", msg);
 }
 
-void complain(const char *errmsg)
+static void
+complain(const char *errmsg)
 {
 	fprintf(stderr, "%s\n", errmsg);
 }
 
 
-/* param str is just a performance enhancement, which saves us re-creation
- * of the string on every call.
+/* rawmsg is, as the name says, the raw message, in case we have
+ * "raw" formatter requested.
  */
-static inline void
-outputEvent(struct json_object *json)
+static void
+outputEvent(struct json_object *json, const char *const rawmsg)
 {
 	char *cstr = NULL;
 	es_str_t *str = NULL;
 
+	if(outfmt == f_raw) {
+		printf("%s\n", rawmsg);
+		return;
+	}
+
 	switch(outfmt) {
 	case f_json:
 		if(!flatTags) {
@@ -98,7 +113,17 @@ outputEvent(struct json_object *json)
 		ln_fmtEventToXML(json, &str);
 		break;
 	case f_csv:
-	ln_fmtEventToCSV(json, &str, encFmt);
+		ln_fmtEventToCSV(json, &str, encFmt);
+		break;
+	case f_raw:
+		fprintf(stderr, "program error: f_raw should not occur "
+			"here (file %s, line %d)\n", __FILE__, __LINE__);
+		abort();
+		break;
+	default:
+		fprintf(stderr, "program error: default case should not occur "
+			"here (file %s, line %d)\n", __FILE__, __LINE__);
+		abort();
 		break;
 	}
 	if (str != NULL)
@@ -120,7 +145,7 @@ eventHasTag(struct json_object *json, const char *tag)
 	
 	if (tag == NULL)
 		return 1;
-	if ((tagbucket = json_object_object_get(json, "event.tags")) != NULL) {
+	if (json_object_object_get_ex(json, "event.tags", &tagbucket)) {
 		if (json_object_get_type(tagbucket) == json_type_array) {
 			for (i = json_object_array_length(tagbucket) - 1; i >= 0; i--) {
 				tagObj = json_object_array_get_idx(tagbucket, i);
@@ -135,35 +160,88 @@ eventHasTag(struct json_object *json, const char *tag)
 	return 0;
 }
 
+static void
+amendLineNbr(json_object *const json, const int line_nbr)
+{
+	
+	if(addErrLineNbr) {
+		struct json_object *jval;
+		jval = json_object_new_int(line_nbr);
+		json_object_object_add(json, "lognormalizer.line_nbr", jval);
+	}
+}
+
+#define DEFAULT_LINE_SIZE (10 * 1024)
+
+static char *
+read_line(FILE *fp)
+{
+	size_t line_capacity = DEFAULT_LINE_SIZE;
+	char *line = NULL;
+	size_t line_len = 0;
+	char ch = 0;
+	do {
+		ch = fgetc(fp);
+		if (ch == EOF) break;
+		if (line == NULL) {
+			line = malloc(line_capacity);
+		} else if (line_len == line_capacity) {
+			line_capacity *= 2;
+			line = realloc(line, line_capacity);
+		}
+		if (line == NULL) {
+			fprintf(stderr, "Couldn't allocate working-buffer for log-line\n");
+			return NULL;
+		}
+		line[line_len++] = ch;
+	} while(ch != '\n');
+
+	if (line != NULL) {
+		line[--line_len] = '\0';
+		if(line_len > 0 && line[line_len - 1] == '\r')
+			line[--line_len] = '\0';
+	}
+	return line;
+}
+
 /* normalize input data
  */
-void
+static void
 normalize(void)
 {
 	FILE *fp = stdin;
-	char buf[10*1024];
+	char *line = NULL;
 	struct json_object *json = NULL;
+	long long unsigned numParsed = 0;
 	long long unsigned numUnparsed = 0;
 	long long unsigned numWrongTag = 0;
 	char *mandatoryTagCstr = NULL;
+	int line_nbr = 0;	/* must be int to keep compatible with older json-c */
 	
 	if (mandatoryTag != NULL) {
 		mandatoryTagCstr = es_str2cstr(mandatoryTag, NULL);
 	}
 
-	while((fgets(buf, sizeof(buf), fp)) != NULL) {
-		buf[strlen(buf)-1] = '\0';
-		if(strlen(buf) > 0 && buf[strlen(buf)-1] == '\r')
-			buf[strlen(buf)-1] = '\0';
-		if(verbose > 0) fprintf(stderr, "To normalize: '%s'\n", buf);
-		ln_normalize(ctx, buf, strlen(buf), &json);
+	while((line = read_line(fp)) != NULL) {
+		++line_nbr;
+		if(verbose > 0) fprintf(stderr, "To normalize: '%s'\n", line);
+		ln_normalize(ctx, line, strlen(line), &json);
 		if(json != NULL) {
 			if(eventHasTag(json, mandatoryTagCstr)) {
-				if( parsedOnly == 1
-						&& json_object_object_get(json, "unparsed-data") != NULL) {
-					numUnparsed++;
+				struct json_object *dummy;
+				const int parsed = !json_object_object_get_ex(json,
+					"unparsed-data", &dummy);
+				if(parsed) {
+					numParsed++;
+					if(recOutput & OUTPUT_PARSED_RECS) {
+						outputEvent(json, line);
+					}
 				} else {
-					outputEvent(json);
+					numUnparsed++;
+					amendLineNbr(json, line_nbr);
+					if(recOutput & OUTPUT_UNPARSED_RECS) {
+						outputEvent(json, line);
+					}
 				}
 			} else {
 				numWrongTag++;
@@ -171,11 +249,16 @@ normalize(void)
 			json_object_put(json);
 			json = NULL;
 		}
+        free(line);
 	}
-	if(numUnparsed > 0)
+	if(outputNbrUnparsed && numUnparsed > 0)
 		fprintf(stderr, "%llu unparsable entries\n", numUnparsed);
 	if(numWrongTag > 0)
 		fprintf(stderr, "%llu entries with wrong tag dropped\n", numWrongTag);
+	if(outputSummaryLine) {
+		fprintf(stderr, "%llu records processed, %llu parsed, %llu unparsed\n",
+			numParsed+numUnparsed, numParsed, numUnparsed);
+	}
 	free(mandatoryTagCstr);
 }
 
@@ -184,19 +267,39 @@ normalize(void)
  * Generate a command file for the GNU DOT tools.
  */
 static void
-genDOT()
+genDOT(void)
 {
 	es_str_t *str;
 
 	str = es_newStr(1024);
-	ln_genDotPTreeGraph(ctx->ptree, &str);
+	ln_genDotPDAGGraph(ctx->pdag, &str);
 	fwrite(es_getBufAddr(str), 1, es_strlen(str), fpDOT);
 }
 
+static
+void printVersion(void)
+{
+	fprintf(stderr, "lognormalizer version: " VERSION "\n");
+	fprintf(stderr, "liblognorm version: %s\n", ln_version());
+	fprintf(stderr, "\tadvanced stats: %s\n",
+		ln_hasAdvancedStats() ? "available" : "not available");
+}
+
 static void
 handle_generic_option(const char* opt) {
 	if (strcmp("allowRegex", opt) == 0) {
-		ln_setCtxOpts(ctx, 1);
+		ln_setCtxOpts(ctx, LN_CTXOPT_ALLOW_REGEX);
+	} else if (strcmp("addExecPath", opt) == 0) {
+		ln_setCtxOpts(ctx, LN_CTXOPT_ADD_EXEC_PATH);
+	} else if (strcmp("addOriginalMsg", opt) == 0) {
+		ln_setCtxOpts(ctx, LN_CTXOPT_ADD_ORIGINALMSG);
+	} else if (strcmp("addRule", opt) == 0) {
+		ln_setCtxOpts(ctx, LN_CTXOPT_ADD_RULE);
+	} else if (strcmp("addRuleLocation", opt) == 0) {
+		ln_setCtxOpts(ctx, LN_CTXOPT_ADD_RULE_LOCATION);
+	} else {
+		fprintf(stderr, "invalid -o option '%s'\n", opt);
+		exit(1);
 	}
 }
 
@@ -205,16 +308,30 @@ static void usage(void)
 fprintf(stderr,
 	"Options:\n"
 	"    -r<rulebase> Rulebase to use. This is required option\n"
-	"    -e<json|xml|csv>\n"
-	"                 Change output format. By default, Mitre CEE is used\n"
+	"    -H           print summary line (nbr of msgs Handled)\n"
+	"    -U           print number of unparsed messages (only if non-zero)\n"
+	"    -e<json|xml|csv|cee-syslog|raw>\n"
+	"                 Change output format. By default, json is used\n"
+	"                 Raw is exactly like the input. It is useful in combination\n"
+	"                 with -p/-P options to extract known good/bad messages\n"
 	"    -E<format>   Encoder-specific format (used for CSV, read docs)\n"
 	"    -T           Include 'event.tags' in JSON format\n"
 	"    -oallowRegex Allow regexp matching (read docs about performance penalty)\n"
+	"    -oaddRule    Add a mockup of the matching rule.\n"
+	"    -oaddRuleLocation Add location of matching rule to metadata\n"
+	"    -oaddExecPath Add exec_path attribute to output\n"
+	"    -oaddOriginalMsg Always add original message to output, not just in error case\n"
 	"    -p           Print back only if the message has been parsed succesfully\n"
+	"    -P           Print back only if the message has NOT been parsed succesfully\n"
+	"    -L           Add source file line number information to unparsed line output\n"
 	"    -t<tag>      Print back only messages matching the tag\n"
-	"    -v           Print debug. When used 3 times, prints parse tree\n"
+	"    -v           Print debug. When used 3 times, prints parse DAG\n"
+	"    -V           Print version information\n"
 	"    -d           Print DOT file to stdout and exit\n"
 	"    -d<filename> Save DOT file to the filename\n"
+	"    -s<filename> Print parse dag statistics and exit\n"
+	"    -S<filename> Print extended parse dag statistics and exit (includes -s)\n"
+	"    -x<filename> Print statistics as dot file (called only)\n"
 	"\n"
 	);
 }
@@ -224,6 +341,9 @@ int main(int argc, char *argv[])
 	int opt;
 	char *repository = NULL;
 	int ret = 0;
+	FILE *fpStats = NULL;
+	FILE *fpStatsDOT = NULL;
+	int extendedStats = 0;
 
 	if((ctx = ln_initCtx()) == NULL) {
 		complain("Could not initialize liblognorm context");
@@ -231,18 +351,51 @@ int main(int argc, char *argv[])
 		goto exit;
 	}
 	
-	while((opt = getopt(argc, argv, "d:e:r:E:vpt:To:h")) != -1) {
+	while((opt = getopt(argc, argv, "d:s:S:e:r:E:vVpPt:To:hHULx:")) != -1) {
 		switch (opt) {
+		case 'V':
+			printVersion();
+			exit(1);
+			break;
 		case 'd': /* generate DOT file */
 			if(!strcmp(optarg, "")) {
 				fpDOT = stdout;
 			} else {
 				if((fpDOT = fopen(optarg, "w")) == NULL) {
+					perror(optarg);
 					complain("Cannot open DOT file");
 					ret = 1;
 					goto exit;
 				}
 			}
+			break;
+		case 'x': /* generate statistics DOT file */
+			if(!strcmp(optarg, "")) {
+				fpStatsDOT = stdout;
+			} else {
+				if((fpStatsDOT = fopen(optarg, "w")) == NULL) {
+					perror(optarg);
+					complain("Cannot open statistics DOT file");
+					ret = 1;
+					goto exit;
+				}
+			}
+			break;
+		case 'S': /* generate pdag statistic file */
+			extendedStats = 1;
+			/* INTENTIONALLY NO BREAK! - KEEP order! */
+		case 's': /* generate pdag statistic file */
+			if(!strcmp(optarg, "-")) {
+				fpStats = stdout;
+			} else {
+				if((fpStats = fopen(optarg, "w")) == NULL) {
+					perror(optarg);
+					complain("Cannot open parser statistics file");
+					ret = 1;
+					goto exit;
+				}
+			}
+			break;
 		case 'v':
 			verbose++;
 			break;
@@ -250,7 +403,19 @@ int main(int argc, char *argv[])
 			encFmt = es_newStrFromCStr(optarg, strlen(optarg));
 			break;
 		case 'p':
-			parsedOnly = 1;
+			recOutput = OUTPUT_PARSED_RECS;
+			break;
+		case 'P':
+			recOutput = OUTPUT_UNPARSED_RECS;
+			break;
+		case 'H':
+			outputSummaryLine = 1;
+			break;
+		case 'U':
+			outputNbrUnparsed = 1;
+			break;
+		case 'L':
+			addErrLineNbr = 1;
 			break;
 		case 'T':
 			flatTags = 1;
@@ -260,8 +425,12 @@ int main(int argc, char *argv[])
 				outfmt = f_json;
 			} else if(!strcmp(optarg, "xml")) {
 				outfmt = f_xml;
+			} else if(!strcmp(optarg, "cee-syslog")) {
+				outfmt = f_syslog;
 			} else if(!strcmp(optarg, "csv")) {
 				outfmt = f_csv;
+			} else if(!strcmp(optarg, "raw")) {
+				outfmt = f_raw;
 			}
 			break;
 		case 'r': /* rule base to use */
@@ -275,6 +444,7 @@ int main(int argc, char *argv[])
 			handle_generic_option(optarg);
 			break;
 		case 'h':
+		default:
 			usage();
 			ret = 1;
 			goto exit;
@@ -308,10 +478,18 @@ int main(int argc, char *argv[])
 		goto exit;
 	}
 
-	if(verbose > 2) ln_displayPTree(ctx->ptree, 0);
+	if(verbose > 2) ln_displayPDAG(ctx);
 
 	normalize();
 
+	if(fpStats != NULL) {
+		ln_fullPdagStats(ctx, fpStats, extendedStats);
+	}
+
+	if(fpStatsDOT != NULL) {
+		ln_fullPDagStatsDOT(ctx, fpStatsDOT);
+	}
+
 exit:
 	if (ctx) ln_exitCtx(ctx);
 	if (encFmt != NULL)
