@@ -1,4 +1,6 @@
 /* samp.c -- code for ln_samp objects.
+ * This code handles rulebase processing. Rulebases have been called
+ * "sample bases" in the early days of liblognorm, thus the name.
  *
  * Copyright 2010-2015 by Rainer Gerhards and Adiscon GmbH.
  *
@@ -29,37 +31,161 @@
 #include <string.h>
 #include <assert.h>
 #include <ctype.h>
+#include <errno.h>
 
-#include "json_compatibility.h"
 #include "liblognorm.h"
 #include "lognorm.h"
 #include "samp.h"
 #include "internal.h"
 #include "parser.h"
+#include "pdag.h"
+#include "v1_liblognorm.h"
+#include "v1_ptree.h"
 
 
-/**
- * Construct a sample object.
- */
-struct ln_samp*
-ln_sampCreate(ln_ctx __attribute__((unused)) ctx)
-{
-	struct ln_samp* samp;
-
-	if((samp = calloc(1, sizeof(struct ln_samp))) == NULL)
-		goto done;
-
-	/* place specific init code here (none at this time) */
-
-done:	return samp;
-}
-
 void
 ln_sampFree(ln_ctx __attribute__((unused)) ctx, struct ln_samp *samp)
 {
 	free(samp);
 }
 
+static int
+ln_parseLegacyFieldDescr(ln_ctx ctx,
+	const char *const buf,
+	const size_t lenBuf,
+	size_t *bufOffs,
+	es_str_t **str,
+	json_object **prscnf)
+{
+	int r = 0;
+	char *cstr;	/* for debug mode strings */
+	char *ftype = NULL;
+	char name[MAX_FIELDNAME_LEN];
+	size_t iDst;
+	struct json_object *json = NULL;
+	char *ed = NULL;
+	es_size_t i = *bufOffs;
+	es_str_t *edata = NULL;
+
+	for(  iDst = 0
+	    ; iDst < (MAX_FIELDNAME_LEN - 1) && i < lenBuf && buf[i] != ':'
+	    ; ++iDst) {
+		name[iDst] = buf[i++];
+	}
+	name[iDst] = '\0';
+	if(iDst == (MAX_FIELDNAME_LEN - 1)) {
+		ln_errprintf(ctx, 0, "field name too long in: %s", buf+(*bufOffs));
+		FAIL(LN_INVLDFDESCR);
+	}
+	if(i == lenBuf) {
+		ln_errprintf(ctx, 0, "field definition wrong in: %s", buf+(*bufOffs));
+		FAIL(LN_INVLDFDESCR);
+	}
+
+	if(iDst == 0) {
+		FAIL(LN_INVLDFDESCR);
+	}
+
+	if(ctx->debug) {
+		ln_dbgprintf(ctx, "parsed field: '%s'", name);
+	}
+
+	if(buf[i] != ':') {
+		ln_errprintf(ctx, 0, "missing colon in: %s", buf+(*bufOffs));
+		FAIL(LN_INVLDFDESCR);
+	}
+	++i; /* skip ':' */
+
+	/* parse and process type (trailing whitespace must be trimmed) */
+	es_emptyStr(*str);
+ 	size_t j = i;
+	/* scan for terminator */
+	while(j < lenBuf && buf[j] != ':' && buf[j] != '{' && buf[j] != '%')
+		++j;
+	/* now trim trailing space backwards */
+	size_t next = j;
+	--j;
+	while(j >= i && isspace(buf[j]))
+		--j;
+	/* now copy */
+	while(i <= j) {
+		CHKR(es_addChar(str, buf[i++]));
+	}
+	/* finally move i to consumed position */
+	i = next;
+
+	if(i == lenBuf) {
+		ln_errprintf(ctx, 0, "premature end (missing %%?) in: %s", buf+(*bufOffs));
+		FAIL(LN_INVLDFDESCR);
+	}
+
+	ftype = es_str2cstr(*str, NULL);
+	ln_dbgprintf(ctx, "field type '%s', i %d", ftype, i);
+
+	if(buf[i] == '{') {
+		struct json_tokener *tokener = json_tokener_new();
+		json = json_tokener_parse_ex(tokener, buf+i, (int) (lenBuf - i));
+		if(json == NULL) {
+			ln_errprintf(ctx, 0, "invalid json in '%s'", buf+i);
+		}
+		i += tokener->char_offset;
+		json_tokener_free(tokener);
+	}
+
+	if(buf[i] == '%') {
+		i++;
+	} else {
+		/* parse extra data */
+		CHKN(edata = es_newStr(8));
+		i++;
+		while(i < lenBuf) {
+			if(buf[i] == '%') {
+				++i;
+				break; /* end of field */
+			}
+			CHKR(es_addChar(&edata, buf[i++]));
+		}
+		es_unescapeStr(edata);
+		if(ctx->debug) {
+			cstr = es_str2cstr(edata, NULL);
+			ln_dbgprintf(ctx, "parsed extra data: '%s'", cstr);
+			free(cstr);
+		}
+	}
+
+	struct json_object *val;
+	*prscnf = json_object_new_object();
+	CHKN(val = json_object_new_string(name));
+	json_object_object_add(*prscnf, "name", val);
+	CHKN(val = json_object_new_string(ftype));
+	json_object_object_add(*prscnf, "type", val);
+	if(edata != NULL) {
+		ed = es_str2cstr(edata, " ");
+		CHKN(val = json_object_new_string(ed));
+		json_object_object_add(*prscnf, "extradata", val);
+	}
+	if(json != NULL) {
+		/* now we need to merge the json params into the main object */
+		struct json_object_iterator it = json_object_iter_begin(json);
+		struct json_object_iterator itEnd = json_object_iter_end(json);
+		while (!json_object_iter_equal(&it, &itEnd)) {
+			struct json_object *const v = json_object_iter_peek_value(&it);
+			json_object_get(v);
+			json_object_object_add(*prscnf, json_object_iter_peek_name(&it), v);
+			json_object_iter_next(&it);
+		}
+	}
+
+	*bufOffs = i;
+done:
+	free(ed);
+	if(edata != NULL)
+		es_deleteStr(edata);
+	free(ftype);
+	if(json != NULL)
+		json_object_put(json);
+	return r;
+}
 
 /**
  * Extract a field description from a sample.
@@ -76,232 +202,81 @@ ln_sampFree(ln_ctx __attribute__((unused)) ctx, struct ln_samp *samp)
  * 		  creation overhead
  * @returns 0 on success, something else otherwise
  */
-static inline int
-addFieldDescr(ln_ctx ctx, struct ln_ptree **subtree, es_str_t *rule,
-	        es_size_t *bufOffs, es_str_t **str)
-{
-	int r;
-	ln_fieldList_t *node = ln_parseFieldDescr(ctx, rule, bufOffs, str, &r);
-	assert(subtree != NULL);
-
-	if (node != NULL) CHKR(ln_addFDescrToPTree(subtree, node));
-done:
-	return r;
-}
-
-ln_fieldList_t*
-ln_parseFieldDescr(ln_ctx ctx, es_str_t *rule, es_size_t *bufOffs, es_str_t **str, int* ret)
+static int
+addFieldDescr(ln_ctx ctx, struct ln_pdag **pdag, es_str_t *rule,
+	        size_t *bufOffs, es_str_t **str)
 {
 	int r = 0;
-	ln_fieldList_t *node;
 	es_size_t i = *bufOffs;
-	char *cstr;	/* for debug mode strings */
-	unsigned char *buf;
+	char *ftype = NULL;
+	const char *buf;
 	es_size_t lenBuf;
-	void* (*constructor_fn)(ln_fieldList_t *, ln_ctx) = NULL;
+	struct json_object *prs_config = NULL;
 
-	buf = es_getBufAddr(rule);
+	buf = (const char*)es_getBufAddr(rule);
 	lenBuf = es_strlen(rule);
 	assert(buf[i] == '%');
 	++i;	/* "eat" ':' */
-	CHKN(node = calloc(1, sizeof(ln_fieldList_t)));
-	node->subtree = NULL;
-	node->next = NULL;
-	node->data = NULL;
-	node->raw_data = NULL;
-	node->parser_data = NULL;
-	node->parser_data_destructor = NULL;
-	CHKN(node->name = es_newStr(16));
 
 	/* skip leading whitespace in field name */
 	while(i < lenBuf && isspace(buf[i]))
 		++i;
-	while(i < lenBuf && buf[i] != ':') {
-		CHKR(es_addChar(&node->name, buf[i++]));
-	}
-
-	if(es_strlen(node->name) == 0) {
-		FAIL(LN_INVLDFDESCR);
-	}
-
-	if(ctx->debug) {
-		cstr = es_str2cstr(node->name, NULL);
-		ln_dbgprintf(ctx, "parsed field: '%s'", cstr);
-		free(cstr);
-	}
-
-	if(buf[i] != ':') {
-		/* may be valid later if we have a loaded CEE dictionary
-		 * and the name is present inside it.
-		 */
-		FAIL(LN_INVLDFDESCR);
-	}
-	++i; /* skip ':' */
-
-	/* parse and process type (trailing whitespace must be trimmed) */
-	es_emptyStr(*str);
- 	size_t j = i;
-	/* scan for terminator */
-	while(j < lenBuf && buf[j] != ':' && buf[j] != '%')
-		++j;
-	/* now trim trailing space backwards */
-	size_t next = j;
-	--j;
-	while(j >= i && isspace(buf[j]))
-		--j;
-	/* now copy */
-	while(i <= j) {
-		CHKR(es_addChar(str, buf[i++]));
-	}
-	/* finally move i to consumed position */
-	i = next;
-
-	if(i == lenBuf) {
-		FAIL(LN_INVLDFDESCR);
-	}
-
-	node->isIPTables = 0; /* first assume no special parser is used */
-	if(!es_strconstcmp(*str, "date-rfc3164")) {
-		node->parser = ln_parseRFC3164Date;
-	} else if(!es_strconstcmp(*str, "date-rfc5424")) {
-		node->parser = ln_parseRFC5424Date;
-	} else if(!es_strconstcmp(*str, "number")) {
-		node->parser = ln_parseNumber;
-	} else if(!es_strconstcmp(*str, "float")) {
-		node->parser = ln_parseFloat;
-	} else if(!es_strconstcmp(*str, "hexnumber")) {
-		node->parser = ln_parseHexNumber;
-	} else if(!es_strconstcmp(*str, "kernel-timestamp")) {
-		node->parser = ln_parseKernelTimestamp;
-	} else if(!es_strconstcmp(*str, "whitespace")) {
-		node->parser = ln_parseWhitespace;
-	} else if(!es_strconstcmp(*str, "ipv4")) {
-		node->parser = ln_parseIPv4;
-	} else if(!es_strconstcmp(*str, "ipv6")) {
-		node->parser = ln_parseIPv6;
-	} else if(!es_strconstcmp(*str, "word")) {
-		node->parser = ln_parseWord;
-	} else if(!es_strconstcmp(*str, "alpha")) {
-		node->parser = ln_parseAlpha;
-	} else if(!es_strconstcmp(*str, "rest")) {
-		node->parser = ln_parseRest;
-	} else if(!es_strconstcmp(*str, "op-quoted-string")) {
-		node->parser = ln_parseOpQuotedString;
-	} else if(!es_strconstcmp(*str, "quoted-string")) {
-		node->parser = ln_parseQuotedString;
-	} else if(!es_strconstcmp(*str, "date-iso")) {
-		node->parser = ln_parseISODate;
-	} else if(!es_strconstcmp(*str, "time-24hr")) {
-		node->parser = ln_parseTime24hr;
-	} else if(!es_strconstcmp(*str, "time-12hr")) {
-		node->parser = ln_parseTime12hr;
-	} else if(!es_strconstcmp(*str, "duration")) {
-		node->parser = ln_parseDuration;
-	} else if(!es_strconstcmp(*str, "cisco-interface-spec")) {
-		node->parser = ln_parseCiscoInterfaceSpec;
-	} else if(!es_strconstcmp(*str, "json")) {
-		node->parser = ln_parseJSON;
-	} else if(!es_strconstcmp(*str, "cee-syslog")) {
-		node->parser = ln_parseCEESyslog;
-	} else if(!es_strconstcmp(*str, "mac48")) {
-		node->parser = ln_parseMAC48;
-	} else if(!es_strconstcmp(*str, "name-value-list")) {
-		node->parser = ln_parseNameValue;
-	} else if(!es_strconstcmp(*str, "cef")) {
-		node->parser = ln_parseCEF;
-	} else if(!es_strconstcmp(*str, "checkpoint-lea")) {
-		node->parser = ln_parseCheckpointLEA;
-	} else if(!es_strconstcmp(*str, "v2-iptables")) {
-		node->parser = ln_parsev2IPTables;
-	} else if(!es_strconstcmp(*str, "iptables")) {
-		node->parser = NULL;
-		node->isIPTables = 1;
-	} else if(!es_strconstcmp(*str, "string-to")) {
-		/* TODO: check extra data!!!! (very important) */
-		node->parser = ln_parseStringTo;
-	} else if(!es_strconstcmp(*str, "char-to")) {
-		/* TODO: check extra data!!!! (very important) */
-		node->parser = ln_parseCharTo;
-	} else if(!es_strconstcmp(*str, "char-sep")) {
-		/* TODO: check extra data!!!! (very important) */
-		node->parser = ln_parseCharSeparated;
-	} else if(!es_strconstcmp(*str, "tokenized")) {
-		node->parser = ln_parseTokenized;
-		constructor_fn = tokenized_parser_data_constructor;
-		node->parser_data_destructor = tokenized_parser_data_destructor;
-	}
-#ifdef FEATURE_REGEXP
-	else if(!es_strconstcmp(*str, "regex")) {
-		node->parser = ln_parseRegex;
-		constructor_fn = regex_parser_data_constructor;
-		node->parser_data_destructor = regex_parser_data_destructor;
-	}
-#endif
-    else if (!es_strconstcmp(*str, "recursive")) {
-        node->parser = ln_parseRecursive;
-        constructor_fn = recursive_parser_data_constructor;
-        node->parser_data_destructor = recursive_parser_data_destructor;
-    } else if (!es_strconstcmp(*str, "descent")) {
-        node->parser = ln_parseRecursive;
-        constructor_fn = descent_parser_data_constructor;
-        node->parser_data_destructor = recursive_parser_data_destructor;
-    } else if (!es_strconstcmp(*str, "interpret")) {
-        node->parser = ln_parseInterpret;
-        constructor_fn = interpret_parser_data_constructor;
-        node->parser_data_destructor = interpret_parser_data_destructor;
-    } else if (!es_strconstcmp(*str, "suffixed")) {
-        node->parser = ln_parseSuffixed;
-        constructor_fn = suffixed_parser_data_constructor;
-        node->parser_data_destructor = suffixed_parser_data_destructor;
-    } else if (!es_strconstcmp(*str, "named_suffixed")) {
-        node->parser = ln_parseSuffixed;
-        constructor_fn = named_suffixed_parser_data_constructor;
-        node->parser_data_destructor = suffixed_parser_data_destructor;
-    } else {
-		cstr = es_str2cstr(*str, NULL);
-		ln_errprintf(ctx, 0, "invalid field type '%s'", cstr);
-		free(cstr);
-		FAIL(LN_INVLDFDESCR);
-	}
-
-	if(buf[i] == '%') {
-		i++;
+	/* check if we have new-style json config */
+	if(buf[i] == '{' || buf[i] == '[') {
+		struct json_tokener *tokener = json_tokener_new();
+		prs_config = json_tokener_parse_ex(tokener, buf+i, (int) (lenBuf - i));
+		i += tokener->char_offset;
+		json_tokener_free(tokener);
+		if(prs_config == NULL || i == lenBuf || buf[i] != '%') {
+			ln_errprintf(ctx, 0, "invalid json in '%s'", buf+i);
+			r = -1;
+			goto done;
+		}
+		*bufOffs = i+1; /* eat '%' - if above ensures it is present */
 	} else {
-		/* parse extra data */
-		CHKN(node->data = es_newStr(8));
-		i++;
-		while(i < lenBuf) {
-			if(buf[i] == '%') {
-				++i;
-				break; /* end of field */
-			}
-			CHKR(es_addChar(&node->data, buf[i++]));
-		}
-		node->raw_data = es_strdup(node->data);
-		es_unescapeStr(node->data);
-		if(ctx->debug) {
-			cstr = es_str2cstr(node->data, NULL);
-			ln_dbgprintf(ctx, "parsed extra data: '%s'", cstr);
-			free(cstr);
-		}
+		*bufOffs = i;
+		CHKR(ln_parseLegacyFieldDescr(ctx, buf, lenBuf, bufOffs, str, &prs_config));
 	}
 
-	if (constructor_fn) node->parser_data = constructor_fn(node, ctx);
+	CHKR(ln_pdagAddParser(ctx, pdag, prs_config));
 
-
-	*bufOffs = i;
 done:
-	if (r != 0) {
-		if (node->name != NULL) es_deleteStr(node->name);
-		free(node);
-		node = NULL;
-	}
-	*ret = r;
-	return node;
+	free(ftype);
+	return r;
+}
+
+
+/**
+ *  Construct a literal parser json definition.
+ */
+static json_object *
+newLiteralParserJSONConf(char lit)
+{
+	char buf[] = "x";
+	buf[0] = lit;
+	struct json_object *val;
+	struct json_object *prscnf = json_object_new_object();
+
+	val = json_object_new_string("literal");
+	json_object_object_add(prscnf, "type", val);
+
+	val = json_object_new_string(buf);
+	json_object_object_add(prscnf, "text", val);
+
+	return prscnf;
 }
 
 /**
  * Parse a Literal string out of the template and add it to the tree.
+ * This function is used to create the unoptimized tree. So we do
+ * one node for each character. These will be compacted by the optimizer
+ * in a later stage. The advantage is that we do not need to care about
+ * splitting the tree. As such the processing is fairly simple:
+ *
+ *   for each character in literal (left-to-right):
+ *      create literal parser object o
+ *      add new DAG node o, advance to it
+ *
  * @param[in] ctx the context
  * @param[in/out] subtree on entry, current subtree, on exist newest
  *    		deepest subtree
@@ -309,22 +284,20 @@ done:
  * @param[in/out] bufOffs parse pointer, up to which offset is parsed
  * 		(is updated so that it points to first char after consumed
  * 		string on exit).
- * @param[out] str literal extracted (is empty, when no litral could be found)
+ * @param    str a work buffer, provided to prevent creation of a new object
  * @return 0 on success, something else otherwise
  */
-static inline int
-parseLiteral(ln_ctx ctx, struct ln_ptree **subtree, es_str_t *rule,
-	     es_size_t *bufOffs, es_str_t **str)
+static int
+parseLiteral(ln_ctx ctx, struct ln_pdag **pdag, es_str_t *rule,
+	     size_t *const __restrict__ bufOffs, es_str_t **str)
 {
 	int r = 0;
-	es_size_t i = *bufOffs;
-	unsigned char *buf;
-	es_size_t lenBuf;
+	size_t i = *bufOffs;
+	unsigned char *buf = es_getBufAddr(rule);
+	const size_t lenBuf = es_strlen(rule);
+	const char *cstr = NULL;
 
 	es_emptyStr(*str);
-	buf = es_getBufAddr(rule);
-	lenBuf = es_strlen(rule);
-	/* extract maximum length literal */
 	while(i < lenBuf) {
 		if(buf[i] == '%') {
 			if(i+1 < lenBuf && buf[i+1] != '%') {
@@ -337,17 +310,26 @@ parseLiteral(ln_ctx ctx, struct ln_ptree **subtree, es_str_t *rule,
 	}
 
 	es_unescapeStr(*str);
+	cstr = es_str2cstr(*str, NULL);
 	if(ctx->debug) {
-		char *cstr = es_str2cstr(*str, NULL);
 		ln_dbgprintf(ctx, "parsed literal: '%s'", cstr);
-		free(cstr);
 	}
 
-	*subtree = ln_buildPTree(*subtree, *str, 0);
 	*bufOffs = i;
+
+	/* we now add the string to the tree */
+	for(i = 0 ; cstr[i] != '\0' ; ++i) {
+		struct json_object *const prscnf = 
+			newLiteralParserJSONConf(cstr[i]);
+		CHKN(prscnf);
+		CHKR(ln_pdagAddParser(ctx, pdag, prscnf));
+	}
+
 	r = 0;
 
-done:	return r;
+done:
+	free((void*)cstr);
+	return r;
 }
 
 
@@ -363,36 +345,39 @@ done:	return r;
  *
  * format: literal1%field:type:extra-data%literal2
  *
- * @returns the new subtree root (or NULL in case of error)
+ * @returns the new dag root (or NULL in case of error)
  */
-static inline int
-addSampToTree(ln_ctx ctx, es_str_t *rule, struct json_object *tagBucket)
+static int
+addSampToTree(ln_ctx ctx,
+	es_str_t *rule,
+	ln_pdag *dag,
+	struct json_object *tagBucket)
 {
 	int r = -1;
-	struct ln_ptree* subtree;
 	es_str_t *str = NULL;
-	es_size_t i;
+	size_t i;
 
-	subtree = ctx->ptree;
 	CHKN(str = es_newStr(256));
 	i = 0;
 	while(i < es_strlen(rule)) {
-		ln_dbgprintf(ctx, "addSampToTree %d of %d", i, es_strlen(rule));
-		CHKR(parseLiteral(ctx, &subtree, rule, &i, &str));
+		LN_DBGPRINTF(ctx, "addSampToTree %zu of %d", i, es_strlen(rule));
+		CHKR(parseLiteral(ctx, &dag, rule, &i, &str));
 		/* After the literal there can be field only*/
 		if (i < es_strlen(rule)) {
-			CHKR(addFieldDescr(ctx, &subtree, rule, &i, &str));
+			CHKR(addFieldDescr(ctx, &dag, rule, &i, &str));
 			if (i == es_strlen(rule)) {
 				/* finish the tree with empty literal to avoid false merging*/
-				CHKR(parseLiteral(ctx, &subtree, rule, &i, &str));
+				CHKR(parseLiteral(ctx, &dag, rule, &i, &str));
 			}
 		}
 	}
 
-	ln_dbgprintf(ctx, "end addSampToTree %d of %d", i, es_strlen(rule));
+	LN_DBGPRINTF(ctx, "end addSampToTree %zu of %d", i, es_strlen(rule));
 	/* we are at the end of rule processing, so this node is a terminal */
-	subtree->flags.isTerminal = 1;
-	subtree->tags = tagBucket;
+	dag->flags.isTerminal = 1;
+	dag->tags = tagBucket;
+	dag->rb_file = strdup(ctx->conf_file);
+	dag->rb_lineno = ctx->conf_ln_nbr;
 
 done:
 	if(str != NULL)
@@ -411,11 +396,11 @@ done:
  * @param[out] str string with "linetype-word" (newly created)
  * @returns 0 on success, something else otherwise
  */
-static inline int
-getLineType(const char *buf, es_size_t lenBuf, es_size_t *offs, es_str_t **str)
+static int
+getLineType(const char *buf, es_size_t lenBuf, size_t *offs, es_str_t **str)
 {
 	int r = -1;
-	es_size_t i;
+	size_t i;
 
 	*str = es_newStr(16);
 	for(i = 0 ; i < lenBuf && buf[i] != '=' ; ++i) {
@@ -441,7 +426,7 @@ done:	return r;
  * 	 	otherwise it is emptied.
  * @returns 0 on success, something else otherwise
  */
-static inline int
+static int
 getPrefix(const char *buf, es_size_t lenBuf, es_size_t offs, es_str_t **str)
 {
 	int r;
@@ -467,7 +452,7 @@ done:	return r;
  * @param[in] offs offset to-be-added text starts
  * @returns 0 on success, something else otherwise
  */
-static inline int
+static int
 extendPrefix(ln_ctx ctx, const char *buf, es_size_t lenBuf, es_size_t offs)
 {
 	return es_addBuf(&ctx->rulePrefix, (char*)buf+offs, lenBuf - offs);
@@ -482,7 +467,7 @@ extendPrefix(ln_ctx ctx, const char *buf, es_size_t lenBuf, es_size_t offs)
  *                       the tagbucket is created if it is NULL
  * @returns 0 on success, something else otherwise
  */
-static inline int
+static int
 addTagStrToBucket(ln_ctx ctx, es_str_t *tagname, struct json_object **tagBucket)
 {
 	int r = -1;
@@ -515,7 +500,7 @@ done:	return r;
  *                       the tagbucket is created if it is NULL
  * @returns 0 on success, something else otherwise
  */
-static inline int
+static int
 processTags(ln_ctx ctx, const char *buf, es_size_t lenBuf, es_size_t *poffs, struct json_object **tagBucket)
 {
 	int r = -1;
@@ -555,8 +540,9 @@ done:	return r;
 }
 
 
+
 /**
- * Process a new rule and add it to tree.
+ * Process a new rule and add it to pdag.
  *
  * @param[in] ctx current context
  * @param[in] buf line buffer
@@ -564,19 +550,18 @@ done:	return r;
  * @param[in] offs offset where rule starts
  * @returns 0 on success, something else otherwise
  */
-static inline int
+static int
 processRule(ln_ctx ctx, const char *buf, es_size_t lenBuf, es_size_t offs)
 {
 	int r = -1;
 	es_str_t *str;
 	struct json_object *tagBucket = NULL;
 
-	ln_dbgprintf(ctx, "sample line to add: '%s'\n", buf+offs);
+	ln_dbgprintf(ctx, "rule line to add: '%s'", buf+offs);
 	CHKR(processTags(ctx, buf, lenBuf, &offs, &tagBucket));
 
 	if(offs == lenBuf) {
-		ln_dbgprintf(ctx, "error, actual message sample part is missing");
-		// TODO: provide some error indicator to app? We definitely must do (a callback?)
+		ln_errprintf(ctx, 0, "error: actual message sample part is missing");
 		goto done;
 	}
 	if(ctx->rulePrefix == NULL) {
@@ -585,7 +570,84 @@ processRule(ln_ctx ctx, const char *buf, es_size_t lenBuf, es_size_t offs)
 		CHKN(str = es_strdup(ctx->rulePrefix));
 	}
 	CHKR(es_addBuf(&str, (char*)buf + offs, lenBuf - offs));
-	addSampToTree(ctx, str, tagBucket);
+	addSampToTree(ctx, str, ctx->pdag, tagBucket);
+	es_deleteStr(str);
+	r = 0;
+done:	return r;
+}
+
+
+static int
+getTypeName(ln_ctx ctx,
+	const char *const __restrict__ buf,
+	const size_t lenBuf,
+	size_t *const __restrict__ offs,
+	char *const __restrict__ dstbuf)
+{
+	int r = -1;
+	size_t iDst;
+	size_t i = *offs;
+	
+	if(buf[i] != '@') {
+		ln_errprintf(ctx, 0, "user-defined type name must "
+			"start with '@'");
+		goto done;
+	}
+	for(  iDst = 0
+	    ; i < lenBuf && buf[i] != ':' && iDst < MAX_TYPENAME_LEN - 1
+	    ; ++i, ++iDst) {
+		if(isspace(buf[i])) {
+			ln_errprintf(ctx, 0, "user-defined type name must "
+				"not contain whitespace");
+			goto done;
+		}
+		dstbuf[iDst] = buf[i];
+	}
+	dstbuf[iDst] = '\0';
+
+	if(i < lenBuf && buf[i] == ':') {
+		r = 0,
+		*offs = i+1; /* skip ":" */
+	}
+done:
+	return r;
+}
+
+/**
+ * Process a type definition and add it to the PDAG
+ * disconnected components. 
+ *
+ * @param[in] ctx current context
+ * @param[in] buf line buffer
+ * @param[in] len length of buffer
+ * @param[in] offs offset where rule starts
+ * @returns 0 on success, something else otherwise
+ */
+static int
+processType(ln_ctx ctx,
+	const char *const __restrict__ buf,
+	const size_t lenBuf,
+	size_t offs)
+{
+	int r = -1;
+	es_str_t *str;
+	char typename[MAX_TYPENAME_LEN];
+
+	ln_dbgprintf(ctx, "type line to add: '%s'", buf+offs);
+	CHKR(getTypeName(ctx, buf, lenBuf, &offs, typename));
+	ln_dbgprintf(ctx, "type name is '%s'", typename);
+
+	ln_dbgprintf(ctx, "type line to add: '%s'", buf+offs);
+	if(offs == lenBuf) {
+		ln_errprintf(ctx, 0, "error: actual message sample part is missing in type def");
+		goto done;
+	}
+	// TODO: optimize
+	CHKN(str = es_newStr(lenBuf));
+	CHKR(es_addBuf(&str, (char*)buf + offs, lenBuf - offs));
+	struct ln_type_pdag *const td = ln_pdagFindType(ctx, typename, 1);
+	CHKN(td);
+	addSampToTree(ctx, str, td->pdag, NULL);
 	es_deleteStr(str);
 	r = 0;
 done:	return r;
@@ -603,7 +665,7 @@ done:	return r;
  * @param [out] strTag obtained tag, if successful
  * @returns 0 on success, something else otherwise
  */
-static inline int
+static int
 getFieldName(ln_ctx __attribute__((unused)) ctx, const char *buf, es_size_t lenBuf, es_size_t *offs, es_str_t **strTag)
 {
 	int r = -1;
@@ -633,7 +695,7 @@ done:	return r;
  * @param[in] len length of buffer
  * @param[in/out] offs on entry: offset first unprocessed position
  */
-static inline void
+static void
 skipWhitespace(ln_ctx __attribute__((unused)) ctx, const char *buf, es_size_t lenBuf, es_size_t *offs)
 {
 	while(*offs < lenBuf && isspace(buf[*offs])) {
@@ -661,7 +723,7 @@ skipWhitespace(ln_ctx __attribute__((unused)) ctx, const char *buf, es_size_t le
  * @param [out] strTag obtained tag, if successful
  * @returns 0 on success, something else otherwise
  */
-static inline int
+static int
 getAnnotationOp(ln_ctx ctx, ln_annot *annot, const char *buf, es_size_t lenBuf, es_size_t *offs)
 {
 	int r = -1;
@@ -723,7 +785,7 @@ fail:	return -1;
  * @param[in] offs offset where annotation starts
  * @returns 0 on success, something else otherwise
  */
-static inline int
+static int
 processAnnotate(ln_ctx ctx, const char *buf, es_size_t lenBuf, es_size_t offs)
 {
 	int r;
@@ -752,15 +814,60 @@ processAnnotate(ln_ctx ctx, const char *buf, es_size_t lenBuf, es_size_t offs)
 done:	return r;
 }
 
-struct ln_samp *
-ln_processSamp(ln_ctx ctx, const char *buf, es_size_t lenBuf)
+/**
+ * Process include directive. This permits to add unlimited layers
+ * of include files.
+ *
+ * @param[in] ctx current context
+ * @param[in] buf line buffer, a C-string
+ * @param[in] offs offset where annotation starts
+ * @returns 0 on success, something else otherwise
+ */
+static int
+processInclude(ln_ctx ctx, const char *buf, const size_t offs)
 {
-	struct ln_samp *samp = NULL;
-    es_str_t *typeStr = NULL;
-    es_size_t offs;
+	int r;
+	const char *const conf_file_save = ctx->conf_file;
+	char *const fname = strdup(buf+offs);
+	size_t lenfname = strlen(fname);
+	const unsigned conf_ln_nbr_save = ctx->conf_ln_nbr;
 
-    if(getLineType(buf, lenBuf, &offs, &typeStr) != 0)
-        goto done;
+	/* trim string - not optimized but also no need to */
+	for(size_t i = lenfname - 1 ; i > 0 ; --i) {
+		if(isspace(fname[i])) {
+			fname[i] = '\0';
+			--lenfname;
+		}
+	}
+
+	CHKR(ln_loadSamples(ctx, fname));
+
+done:
+	free(fname);
+	ctx->conf_file = conf_file_save;
+	ctx->conf_ln_nbr = conf_ln_nbr_save;
+
+	return r;
+}
+
+/**
+ * Reads a rule (sample) stored in buffer buf and creates a new ln_samp object
+ * out of it, which it adds to the pdag (if required).
+ *
+ * @param[ctx] ctx current library context
+ * @param[buf] cstr buffer containing the string contents of the sample
+ * @param[lenBuf] length of the sample contained within buf
+ * @return standard error code
+ */
+static int
+ln_processSamp(ln_ctx ctx, const char *buf, const size_t lenBuf)
+{
+	int r = 0;
+	es_str_t *typeStr = NULL;
+	size_t offs;
+
+	if(getLineType(buf, lenBuf, &offs, &typeStr) != 0)
+		goto done;
 
 	if(!es_strconstcmp(typeStr, "prefix")) {
 		if(getPrefix(buf, lenBuf, offs, &ctx->rulePrefix) != 0) goto done;
@@ -768,13 +875,16 @@ ln_processSamp(ln_ctx ctx, const char *buf, es_size_t lenBuf)
 		if(extendPrefix(ctx, buf, lenBuf, offs) != 0) goto done;
 	} else if(!es_strconstcmp(typeStr, "rule")) {
 		if(processRule(ctx, buf, lenBuf, offs) != 0) goto done;
+	} else if(!es_strconstcmp(typeStr, "type")) {
+		if(processType(ctx, buf, lenBuf, offs) != 0) goto done;
 	} else if(!es_strconstcmp(typeStr, "annotate")) {
 		if(processAnnotate(ctx, buf, lenBuf, offs) != 0) goto done;
+	} else if(!es_strconstcmp(typeStr, "include")) {
+		CHKR(processInclude(ctx, buf, offs));
 	} else {
-		/* TODO error reporting */
 		char *str;
 		str = es_str2cstr(typeStr, NULL);
-		ln_dbgprintf(ctx, "invalid record type detected: '%s'", str);
+		ln_errprintf(ctx, 0, "invalid record type detected: '%s'", str);
 		free(str);
 		goto done;
 	}
@@ -782,8 +892,20 @@ ln_processSamp(ln_ctx ctx, const char *buf, es_size_t lenBuf)
 done:
 	if(typeStr != NULL)
 		es_deleteStr(typeStr);
+	return r;
+}
 
-	return samp;
+
+/* note: comments are only supported at beginning of line! */
+/* skip to end of line */
+void
+ln_sampSkipCommentLine(ln_ctx ctx, FILE * const __restrict__ repo)
+{
+	int c;
+	do {
+		c = fgetc(repo);
+	} while(c != EOF && c != '\n');
+	++ctx->conf_ln_nbr;
 }
 
 
@@ -793,24 +915,45 @@ done:
  * is otherwise hard to debug, so let's see if it is the case...
  * @return 1 if this is a runaway rule, 0 if not
  */
-static int
-chkRunawayRule(ln_ctx ctx, FILE *const __restrict__ repo, const int linenbr)
+int
+ln_sampChkRunawayRule(ln_ctx ctx, FILE *const __restrict__ repo)
 {
 	int r = 1;
 	fpos_t fpos;
 	char buf[6];
+	int cont = 1;
+	int read;
 
 	fgetpos(repo, &fpos);
-	if(fread(buf, sizeof(char), sizeof(buf)-1, repo) != 5)
-		goto done;
-	buf[5] = '\0';
-	if(!strcmp(buf, "rule=")) {
-		ln_errprintf(ctx, 0, "line %d has 'rule=' at begin of line, which "
-			"does look like a typo in the previous lines (unmatched "
-			"%% character) and is forbidden. If valid, please re-foramt "
-			"the rule to start with other characters. Rule ignored.",
-			linenbr);
-		goto done;
+	while(cont) {
+		fpos_t inner_fpos;
+		fgetpos(repo, &inner_fpos);
+		if((read = fread(buf, sizeof(char), sizeof(buf)-1, repo)) == 0) {
+			r = 0;
+			goto done;
+		}
+		if(buf[0] == '\n') {
+			fsetpos(repo, &inner_fpos);
+			fread(buf, sizeof(char), 1, repo); /* skip '\n' */
+			continue;
+		} else if(buf[0] == '#') {
+			fsetpos(repo, &inner_fpos);
+			const unsigned conf_ln_nbr_save = ctx->conf_ln_nbr;
+			ln_sampSkipCommentLine(ctx, repo);
+			ctx->conf_ln_nbr = conf_ln_nbr_save;
+			continue;
+		}
+		if(read != 5)
+			goto done; /* cannot be a rule= line! */
+		cont = 0; /* no comment, so we can decide */
+		buf[5] = '\0';
+		if(!strncmp(buf, "rule=", 5)) {
+			ln_errprintf(ctx, 0, "line has 'rule=' at begin of line, which "
+				"does look like a typo in the previous lines (unmatched "
+				"%% character) and is forbidden. If valid, please re-format "
+				"the rule to start with other characters. Rule ignored.");
+			goto done;
+		}
 	}
 
 	r = 0;
@@ -818,13 +961,26 @@ done:
 	fsetpos(repo, &fpos);
 	return r;
 }
-struct ln_samp *
+
+
+/**
+ * Read a rule (sample) from repository (sequentially).
+ *
+ * Reads a sample starting with the current file position and
+ * creates a new ln_samp object out of it, which it adds to the
+ * pdag.
+ *
+ * @param[in] ctx current library context
+ * @param[in] repo repository descriptor
+ * @param[out] isEof must be set to 0 on entry and is switched to 1 if EOF occured.
+ * @return standard error code
+ */
+static int
 ln_sampRead(ln_ctx ctx, FILE *const __restrict__ repo, int *const __restrict__ isEof)
 {
-	struct ln_samp *samp = NULL;
-	char buf[10*1024]; /**< max size of rule - TODO: make configurable */
+	int r = 0;
+	char buf[64*1024]; /**< max size of rule - TODO: make configurable */
 
-	int linenbr = 1;
 	size_t i = 0;
 	int inParser = 0;
 	int done = 0;
@@ -837,9 +993,9 @@ ln_sampRead(ln_ctx ctx, FILE *const __restrict__ repo, int *const __restrict__ i
 			else
 				done = 1; /* last line missing LF, still process it! */
 		} else if(c == '\n') {
-			++linenbr;
+			++ctx->conf_ln_nbr;
 			if(inParser) {
-				if(chkRunawayRule(ctx, repo, linenbr)) {
+				if(ln_sampChkRunawayRule(ctx, repo)) {
 					/* ignore previous rule */
 					inParser = 0;
 					i = 0;
@@ -848,29 +1004,136 @@ ln_sampRead(ln_ctx ctx, FILE *const __restrict__ repo, int *const __restrict__ i
 			if(!inParser && i != 0)
 				done = 1;
 		} else if(c == '#' && i == 0) {
-			/* note: comments are only supported at beginning of line! */
-			/* skip to end of line */
-			do {
-				c = fgetc(repo);
-			} while(c != EOF && c != '\n');
-			++linenbr;
+			ln_sampSkipCommentLine(ctx, repo);
 			i = 0; /* back to beginning */
 		} else {
 			if(c == '%')
 				inParser = (inParser) ? 0 : 1;
 			buf[i++] = c;
 			if(i >= sizeof(buf)) {
-				ln_errprintf(ctx, 0, "line %d is too long", linenbr);
-				ln_dbgprintf(ctx, "line %d is too long", linenbr);
+				ln_errprintf(ctx, 0, "line is too long");
 				goto done;
 			}
 		}
 	}
 	buf[i] = '\0';
 
-	ln_dbgprintf(ctx, "read sample line: '%s'", buf);
-	ln_processSamp(ctx, buf, i);
+	ln_dbgprintf(ctx, "read rulebase line[~%d]: '%s'", ctx->conf_ln_nbr, buf);
+	CHKR(ln_processSamp(ctx, buf, i));
 
 done:
-	return samp;
+	return r;
 }
+
+/* check rulebase format version. Returns 2 if this is v2 rulebase,
+ * 1 for any pre-v2 and -1 if there was a problem reading the file.
+ */
+static int
+checkVersion(FILE *const fp)
+{
+	char buf[64];
+
+	if(fgets(buf, sizeof(buf), fp) == NULL)
+		return -1;
+	if(!strcmp(buf, "version=2\n")) {
+		return 2;
+	} else {
+		return 1;
+	}
+}
+
+/* we have a v1 rulebase, so let's do all stuff that we need
+ * to make that ole piece of ... work.
+ */
+static int
+doOldCruft(ln_ctx ctx, const char *file)
+{
+	int r = -1;
+	if((ctx->ptree = ln_newPTree(ctx, NULL)) == NULL) {
+		free(ctx);
+		r = -1;
+		goto done;
+	}
+	r = ln_v1_loadSamples(ctx, file);
+done:
+	return r;
+}
+
+/* try to open a rulebase file. This also tries to see if we need to
+ * load it from some pre-configured alternative location.
+ * @returns open file pointer or NULL in case of error
+ */
+static FILE *
+tryOpenRBFile(ln_ctx ctx, const char *const file)
+{
+	FILE *repo = NULL;
+
+	if((repo = fopen(file, "r")) != NULL)
+		goto done;
+	const int eno1 = errno;
+
+	const char *const rb_lib = getenv("LIBLOGNORM_RULEBASES");
+	if(rb_lib == NULL || *file == '/') {
+		ln_errprintf(ctx, eno1, "cannot open rulebase '%s'", file);
+		goto done;
+	}
+
+	char *fname = NULL;
+	asprintf(&fname, (rb_lib[strlen(rb_lib)-1] == '/') ? "%s%s" : "%s/%s", rb_lib, file);
+	if((repo = fopen(fname, "r")) == NULL) {
+		const int eno2 = errno;
+		ln_errprintf(ctx, eno1, "cannot open rulebase '%s'", file);
+		ln_errprintf(ctx, eno2, "also tried to locate %s via "
+			"rulebase directory without success. Expanded "
+			"name was '%s'", file, fname);
+	}
+	free(fname);
+
+done:
+	return repo;
+}
+
+/* @return 0 if all is ok, 1 if an error occured */
+int
+ln_sampLoad(ln_ctx ctx, const char *file)
+{
+	int r = 1;
+	FILE *repo;
+	int isEof = 0;
+
+	ln_dbgprintf(ctx, "loading rulebase file '%s'", file);
+	if(file == NULL) goto done;
+	if((repo = tryOpenRBFile(ctx, file)) == NULL)
+		goto done;
+	const int version = checkVersion(repo);
+	ln_dbgprintf(ctx, "rulebase version is %d\n", version);
+	if(version == -1) {
+		ln_errprintf(ctx, errno, "error determing version of %s", file);
+		goto done;
+	}
+	if(ctx->version != 0 && version != ctx->version) {
+		ln_errprintf(ctx, errno, "rulebase '%s' must be version %d, but is version %d "
+			" - can not be processed", file, ctx->version, version);
+		goto done;
+	}
+	ctx->version = version;
+	if(ctx->version == 1) {
+		fclose(repo);
+		r = doOldCruft(ctx, file);
+		goto done;
+	}
+
+	/* now we are in our native code */
+	++ctx->conf_ln_nbr; /* "version=2" is line 1! */
+	while(!isEof) {
+		CHKR(ln_sampRead(ctx, repo, &isEof));
+	}
+	fclose(repo);
+	r = 0;
+
+	if(ctx->include_level == 1)
+		ln_pdagOptimize(ctx);
+done:
+	return r;
+}
+
