@@ -1,3 +1,7 @@
+# environment variables:
+# GREP - if set, can be used to use alternative grep version
+#        Most important use case is to use GNU grep (ggrep)
+#        on Solaris. If unset, use "grep".
 set -e
 
 if [ "x$debug" == "xon" ]; then #get core-dump on crash
@@ -32,7 +36,10 @@ execute() {
 }
 
 assert_output_contains() {
-    cat test.out | grep -F "$1"
+    if [ "x$GREP" == "x" ]; then
+       GREP=grep
+    fi
+    cat test.out | $GREP -F "$1"
 }
 
 assert_output_json_eq() {
@@ -48,15 +55,21 @@ rulebase_file_name() {
 }
 
 reset_rules() {
-    local rb_file=$(rulebase_file_name $1)
+    rb_file=$(rulebase_file_name $1)
     rm -f $rb_file
 }
 
 add_rule() {
-    local rb_file=$(rulebase_file_name $2)
+    rb_file=$(rulebase_file_name $2)
     echo $1 >> $rb_file
 }
 
+add_rule_no_LF() {
+    rb_file=$(rulebase_file_name $2)
+    echo -n $1 >> $rb_file
+}
+
+
 cleanup_tmp_files() {
     rm -f test.out *.rulebase 
 }
